@@ -15,6 +15,9 @@
  */
 package com.googlecode.jtype;
 
+import static com.googlecode.jtype.Utils.checkNotNull;
+import static com.googlecode.jtype.Utils.checkTrue;
+
 import java.io.Serializable;
 import java.lang.reflect.GenericArrayType;
 import java.lang.reflect.MalformedParameterizedTypeException;
@@ -27,14 +30,14 @@ import java.util.HashMap;
 import java.util.HashSet;
 import java.util.LinkedHashMap;
 import java.util.Map;
-import java.util.Set;
 import java.util.Map.Entry;
+import java.util.Set;
 
 /**
  * Provides utility methods for working with types.
  * 
  * @author Mark Hobson
- * @version $Id: TypeUtils.java 49 2009-10-07 17:10:57Z markhobson $
+ * @version $Id: TypeUtils.java 113 2011-11-25 18:14:23Z markhobson@gmail.com $
  */
 public final class TypeUtils
 {
@@ -68,10 +71,92 @@ public final class TypeUtils
 	
 	// public methods ---------------------------------------------------------
 	
+	public static void accept(Type type, TypeVisitor visitor)
+	{
+		checkNotNull(type, "type");
+		checkNotNull(visitor, "visitor");
+		
+		if (type instanceof Class<?>)
+		{
+			visitor.visit((Class<?>) type);
+		}
+		else if (type instanceof TypeVariable<?>)
+		{
+			TypeVariable<?> typeVariable = (TypeVariable<?>) type;
+			
+			if (visitor.beginVisit(typeVariable))
+			{
+				// visit bounds
+			
+				Type[] bounds = typeVariable.getBounds();
+				
+				for (int i = 0; i < bounds.length; i++)
+				{
+					visitor.visitTypeVariableBound(bounds[i], i);
+				}
+			}
+			
+			visitor.endVisit(typeVariable);
+		}
+		else if (type instanceof GenericArrayType)
+		{
+			visitor.visit((GenericArrayType) type);
+		}
+		else if (type instanceof ParameterizedType)
+		{
+			ParameterizedType parameterizedType = (ParameterizedType) type;
+			
+			if (visitor.beginVisit(parameterizedType))
+			{
+				// visit actual type arguments
+				
+				Type[] actualTypeArguments = parameterizedType.getActualTypeArguments();
+				
+				for (int i = 0; i < actualTypeArguments.length; i++)
+				{
+					visitor.visitActualTypeArgument(actualTypeArguments[i], i);
+				}
+			}
+			
+			visitor.endVisit(parameterizedType);
+		}
+		else if (type instanceof WildcardType)
+		{
+			WildcardType wildcardType = (WildcardType) type;
+
+			if (visitor.beginVisit(wildcardType))
+			{
+				// visit upper bounds
+				
+				Type[] upperBounds = wildcardType.getUpperBounds();
+				
+				for (int i = 0; i < upperBounds.length; i++)
+				{
+					visitor.visitUpperBound(upperBounds[i], i);
+				}
+				
+				// visit lower bounds
+				
+				Type[] lowerBounds = wildcardType.getLowerBounds();
+				
+				for (int i = 0; i < lowerBounds.length; i++)
+				{
+					visitor.visitLowerBound(lowerBounds[i], i);
+				}
+			}
+				
+			visitor.endVisit(wildcardType);
+		}
+		else
+		{
+			throw new IllegalArgumentException("Unknown type: " + type);
+		}
+	}
+	
 	public static boolean isAssignable(Type supertype, Type type)
 	{
-		Utils.checkNotNull(supertype, "supertype");
-		Utils.checkNotNull(type, "type");
+		checkNotNull(supertype, "supertype");
+		checkNotNull(type, "type");
 		
 		if (supertype.equals(type))
 		{
@@ -105,6 +190,11 @@ public final class TypeUtils
 				return isArraySupertype((Class<?>) supertype);
 			}
 			
+			if (type instanceof WildcardType)
+			{
+				return isClassAssignableToWildcardType((Class<?>) supertype, (WildcardType) type);
+			}
+			
 			return false;
 		}
 		
@@ -194,7 +284,7 @@ public final class TypeUtils
 	
 	public static Class<?> getErasedReferenceType(Type type)
 	{
-		Utils.checkTrue(isReferenceType(type), "type is not a reference type: ", type);
+		checkTrue(isReferenceType(type), "type is not a reference type: ", type);
 		
 		return (Class<?>) getErasedType(type);
 	}
@@ -214,6 +304,11 @@ public final class TypeUtils
 			|| (type instanceof GenericArrayType);
 	}
 	
+	public static boolean isPrimitive(Type type)
+	{
+		return (type instanceof Class<?>) && ((Class<?>) type).isPrimitive();
+	}
+	
 	public static Type getComponentType(Type type)
 	{
 		if (type instanceof Class<?>)
@@ -233,7 +328,7 @@ public final class TypeUtils
 	
 	public static Type getArrayType(Type componentType)
 	{
-		Utils.checkNotNull(componentType, "componentType");
+		checkNotNull(componentType, "componentType");
 		
 		if (componentType instanceof Class<?>)
 		{
@@ -243,53 +338,68 @@ public final class TypeUtils
 		return Types.genericArrayType(componentType);
 	}
 	
+	public static boolean isParameterizedType(Type type, Class<?> rawType)
+	{
+		checkNotNull(type, "type");
+		checkNotNull(rawType, "rawType");
+		
+		ParameterizedType parameterizedType;
+		
+		try
+		{
+			parameterizedType = Types.unboundedParameterizedType(rawType);
+		}
+		catch (MalformedParameterizedTypeException exception)
+		{
+			return false;
+		}
+		
+		return isAssignable(parameterizedType, type);
+	}
+
+	/**
+	 * @deprecated Use {@link #isParameterizedType(Type, Class)} instead.
+	 */
+	@Deprecated
 	public static boolean isSimpleParameterizedType(Type type, Class<?> rawType)
 	{
-		Utils.checkNotNull(type, "type");
-		Utils.checkNotNull(rawType, "rawType");
-		
-		if (!(type instanceof ParameterizedType))
+		if (!isParameterizedType(type, rawType))
 		{
 			return false;
 		}
 		
-		ParameterizedType paramType = (ParameterizedType) type;
-		
-		Type paramRawType = paramType.getRawType();
-		
-		if (!(paramRawType instanceof Class<?>))
-		{
-			return false;
-		}
-		
-		Class<?> paramRawClass = (Class<?>) paramRawType;
-		
-		if (!rawType.isAssignableFrom(paramRawClass))
-		{
-			return false;
-		}
-		
-		Type[] typeArgs = paramType.getActualTypeArguments();
+		Type[] typeArgs = ((ParameterizedType) type).getActualTypeArguments();
 		
 		return (typeArgs.length == 1);
 	}
 	
+	/**
+	 * @deprecated Use {@link #getActualTypeArgument(Type, int)} instead.
+	 */
+	@Deprecated
 	public static Type getActualTypeArgument(Type type)
 	{
-		Utils.checkNotNull(type, "type");
+		return getActualTypeArgument(type, 0);
+	}
+	
+	public static Type getActualTypeArgument(Type type, int argIndex)
+	{
+		checkNotNull(type, "type");
+		checkTrue(type instanceof ParameterizedType, "type must be a ParameterizedType: ", type);
 		
 		ParameterizedType paramType = (ParameterizedType) type;
 		
 		Type[] typeArgs = paramType.getActualTypeArguments();
 		
-		Utils.checkTrue(typeArgs.length == 1, "type must be a ParameterizedType with one actual type argument: ", type);
+		checkTrue(argIndex >= 0, "argIndex must be non-negative: ", argIndex);
+		checkTrue(argIndex < typeArgs.length, "argIndex must be less than the number of type parameters: ", argIndex);
 		
-		return typeArgs[0];
+		return typeArgs[argIndex];
 	}
 	
 	public static Type getResolvedSuperclass(Type type)
 	{
-		Utils.checkNotNull(type, "type");
+		checkNotNull(type, "type");
 		
 		Class<?> rawType = getErasedReferenceType(type);
 		Type supertype = rawType.getGenericSuperclass();
@@ -304,7 +414,7 @@ public final class TypeUtils
 	
 	public static Type[] getResolvedInterfaces(Type type)
 	{
-		Utils.checkNotNull(type, "type");
+		checkNotNull(type, "type");
 		
 		Class<?> rawType = getErasedReferenceType(type);
 		Type[] interfaces = rawType.getGenericInterfaces();
@@ -318,6 +428,18 @@ public final class TypeUtils
 		return resolvedInterfaces;
 	}
 	
+	public static <T> Type getResolvedSupertype(Class<? extends T> type, Class<T> rawSupertype)
+	{
+		Type resolvedSupertype = getResolvedSupertype((Type) type, rawSupertype);
+		
+		if (resolvedSupertype == null)
+		{
+			throw new IllegalStateException("type must extend or implement supertype: " + type.getName());
+		}
+
+		return resolvedSupertype;
+	}
+	
 	public static String toString(Type type)
 	{
 		return toString(type, ClassSerializers.QUALIFIED);
@@ -325,39 +447,16 @@ public final class TypeUtils
 	
 	public static String toString(Type type, ClassSerializer serializer)
 	{
-		if (type instanceof Class<?>)
+		if (type == null)
 		{
-			Class<?> klass = (Class<?>) type;
-			
-			if (klass.isArray())
-			{
-				return toString(klass.getComponentType(), serializer) + "[]";
-			}
-			
-			return serializer.toString(klass);
+			return String.valueOf(type);
 		}
 		
-		if (type instanceof TypeVariable<?>)
-		{
-			return DefaultTypeVariable.toString((TypeVariable<?>) type, serializer);
-		}
+		SerializingTypeVisitor visitor = new SerializingTypeVisitor(serializer);
 		
-		if (type instanceof GenericArrayType)
-		{
-			return DefaultGenericArrayType.toString((GenericArrayType) type, serializer);
-		}
+		accept(type, visitor);
 		
-		if (type instanceof ParameterizedType)
-		{
-			return DefaultParameterizedType.toString((ParameterizedType) type, serializer);
-		}
-		
-		if (type instanceof WildcardType)
-		{
-			return DefaultWildcardType.toString((WildcardType) type, serializer);
-		}
-		
-		return String.valueOf(type);
+		return visitor.toString();
 	}
 
 	public static String toUnqualifiedString(Type type)
@@ -370,23 +469,6 @@ public final class TypeUtils
 		return toString(type, ClassSerializers.SIMPLE);
 	}
 	
-	// package methods --------------------------------------------------------
-	
-	static StringBuilder appendBounds(StringBuilder builder, Type[] bounds, ClassSerializer serializer)
-	{
-		for (int i = 0; i < bounds.length; i++)
-		{
-			if (i > 0)
-			{
-				builder.append(" & ");
-			}
-			
-			builder.append(toString(bounds[i], serializer));
-		}
-		
-		return builder;
-	}
-	
 	// private methods --------------------------------------------------------
 	
 	private static void putPrimitiveSubtypes(Map<Class<?>, Set<Class<?>>> subtypesByPrimitive, Class<?> primitiveType,
@@ -414,6 +496,19 @@ public final class TypeUtils
 		return supertype.isAssignableFrom(type);
 	}
 	
+	private static boolean isClassAssignableToWildcardType(Class<?> supertype, WildcardType type)
+	{
+		for (Type upperBound : type.getUpperBounds())
+		{
+			if (!isAssignable(supertype, upperBound))
+			{
+				return false;
+			}
+		}
+		
+		return true;
+	}
+	
 	private static boolean isParameterizedTypeAssignable(ParameterizedType supertype, ParameterizedType type)
 	{
 		Type rawSupertype = supertype.getRawType();
@@ -632,7 +727,7 @@ public final class TypeUtils
 	
 	private static <K, V> Map<K, V> normalize(Map<K, V> map)
 	{
-		// TODO: will this cause an infinite look with recursive bounds?
+		// TODO: will this cause an infinite loop with recursive bounds?
 		
 		for (Entry<K, V> entry : map.entrySet())
 		{
@@ -649,4 +744,37 @@ public final class TypeUtils
 		
 		return map;
 	}
+	
+	private static Type getResolvedSupertype(Type type, Class<?> rawSupertype)
+	{
+		Class<?> rawCurrentType = TypeUtils.getErasedReferenceType(type);
+		
+		if (rawSupertype.equals(rawCurrentType))
+		{
+			return type;
+		}
+		
+		// try interfaces
+		
+		for (Type interfaceType : TypeUtils.getResolvedInterfaces(type))
+		{
+			Type resolvedType = getResolvedSupertype(interfaceType, rawSupertype);
+			
+			if (resolvedType != null)
+			{
+				return resolvedType;
+			}
+		}
+		
+		// try superclass
+		
+		Type supertype = TypeUtils.getResolvedSuperclass(type);
+		
+		if (supertype == null)
+		{
+			return null;
+		}
+		
+		return getResolvedSupertype(supertype, rawSupertype);
+	}
 }
