@@ -48,11 +48,14 @@ import org.junit.Test;
  * Tests {@code TypeUtils}.
  * 
  * @author Mark Hobson
- * @version $Id: TypeUtilsTest.java 48 2009-10-02 16:57:38Z markhobson $
+ * @version $Id: TypeUtilsTest.java 114 2011-11-25 18:15:48Z markhobson@gmail.com $
  * @see TypeUtils
  */
 public class TypeUtilsTest
 {
+	// TODO: split into smaller test cases
+	// TODO: use local types rather than JDK classes in tests
+	
 	// classes ----------------------------------------------------------------
 	
 	private static class IntegerArrayList extends ArrayList<Integer>
@@ -466,6 +469,30 @@ public class TypeUtilsTest
 		assertAsymmetricallyAssignable(valueOf("List<?>"), valueOf("List<Integer>"));
 	}
 	
+	/**
+	 * Tests that parameterized type upper bounded wildcard type arguments are assignable to wildcard types.
+	 * 
+	 * {@literal List<?> <: List<? extends Number>}
+	 */
+	@Test
+	public void isAssignableWithWildcardParameterizedTypeFromUpperBoundedWildcardParameterizedType()
+	{
+		assertAsymmetricallyAssignable(valueOf("List<?>"), valueOf("List<? extends Number>"));
+	}
+	
+	/**
+	 * Tests that parameterized type lower bounded wildcard type arguments are assignable to wildcard types.
+	 * 
+	 * {@literal List<?> <: List<? super Number>}
+	 */
+	// TODO: fix
+	@Ignore
+	@Test
+	public void isAssignableWithWildcardParameterizedTypeFromLowerBoundedWildcardParameterizedType()
+	{
+		assertAsymmetricallyAssignable(valueOf("List<?>"), valueOf("List<? super Number>"));
+	}
+	
 	/**
 	 * Tests that parameterized type arguments are assignable to wildcard types on their upper bound.
 	 * 
@@ -503,8 +530,6 @@ public class TypeUtilsTest
 	/**
 	 * {@literal List<? extends Number> <: List<? extends Integer>}
 	 */
-	// TODO: fix
-	@Ignore
 	@Test
 	public void isAssignableWithUpperBoundedWildcardParameterizedTypeFromInBoundsUpperBoundedWildcardParameterizedType()
 	{
@@ -665,6 +690,94 @@ public class TypeUtilsTest
 		assertAssignable(Collection.class, type);
 	}
 	
+	/**
+	 * Tests that unbounded wildcards are assignable to Object.
+	 * 
+	 * {@literal Object <: ?}
+	 */
+	@Test
+	public void isAssignableWithObjectFromUnboundedWildcardType()
+	{
+		assertAssignable(Object.class, Types.unboundedWildcardType());
+	}
+	
+	/**
+	 * Tests that upper bounded wildcards are assignable to their upper bound.
+	 * 
+	 * {@literal Number <: ? extends Number}
+	 */
+	@Test
+	public void isAssignableWithBoundFromUpperBoundedWildcardType()
+	{
+		assertAssignable(Number.class, Types.upperBoundedWildcardType(Number.class));
+	}
+	
+	/**
+	 * Tests that upper bounded wildcards are assignable to supertypes of their upper bound.
+	 * 
+	 * {@literal Number <: ? extends Integer}
+	 */
+	@Test
+	public void isAssignableWithBoundSupertypeFromUpperBoundedWildcardType()
+	{
+		assertAssignable(Number.class, Types.upperBoundedWildcardType(Integer.class));
+	}
+	
+	/**
+	 * Tests that upper bounded wildcards are not assignable to subtypes of their upper bound.
+	 * 
+	 * {@literal Integer !<: ? extends Number}
+	 */
+	@Test
+	public void isAssignableWithBoundSubtypeFromUpperBoundedWildcardType()
+	{
+		assertUnassignable(Integer.class, Types.upperBoundedWildcardType(Number.class));
+	}
+	
+	/**
+	 * Tests that lower bounded wildcards are assignable to Object.
+	 * 
+	 * {@literal Object <: ? super Number}
+	 */
+	@Test
+	public void isAssignableWithObjectFromLowerBoundedWildcardType()
+	{
+		assertAssignable(Object.class, Types.lowerBoundedWildcardType(Number.class));
+	}
+	
+	/**
+	 * Tests that lower bounded wildcards are not assignable to their lower bound.
+	 * 
+	 * {@literal Number !<: ? super Number}
+	 */
+	@Test
+	public void isAssignableWithBoundFromLowerBoundedWildcardType()
+	{
+		assertUnassignable(Number.class, Types.lowerBoundedWildcardType(Number.class));
+	}
+	
+	/**
+	 * Tests that lower bounded wildcards are not assignable to supertypes of their lower bound.
+	 * 
+	 * {@literal Number !<: ? super Integer}
+	 */
+	@Test
+	public void isAssignableWithBoundSupertypeFromLowerBoundedWildcardType()
+	{
+		assertUnassignable(Number.class, Types.lowerBoundedWildcardType(Integer.class));
+	}
+	
+	/**
+	 * Tests that upper bounded wildcards are not assignable to subtypes of their upper bound.
+	 * 
+	 * {@literal Integer <: ? super Number}
+	 */
+	@Test
+	public void isAssignableWithBoundSubtypeFromLowerBoundedWildcardType()
+	{
+		assertUnassignable(Integer.class, Types.lowerBoundedWildcardType(Number.class));
+	}
+	
 	// isInstance tests -------------------------------------------------------
 	
 	@Test(expected = NullPointerException.class)
@@ -948,6 +1061,58 @@ public class TypeUtilsTest
 		assertFalse(TypeUtils.isArray(Types.unboundedWildcardType()));
 	}
 	
+	// isPrimitive tests ------------------------------------------------------
+	
+	@Test
+	public void isPrimitiveWithNull()
+	{
+		assertFalse(TypeUtils.isPrimitive(null));
+	}
+	
+	@Test
+	public void isPrimitiveWithPrimitives()
+	{
+		assertTrue(TypeUtils.isPrimitive(Boolean.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Character.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Byte.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Short.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Integer.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Long.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Float.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Double.TYPE));
+		assertTrue(TypeUtils.isPrimitive(Void.TYPE));
+	}
+	
+	@Test
+	public void isPrimitiveWithClass()
+	{
+		assertFalse(TypeUtils.isPrimitive(Integer.class));
+	}
+	
+	@Test
+	public void isPrimitiveWithTypeVariable()
+	{
+		assertFalse(TypeUtils.isPrimitive(Types.typeVariable(declaration, "T")));
+	}
+	
+	@Test
+	public void isPrimitiveWithGenericArrayType()
+	{
+		assertFalse(TypeUtils.isPrimitive(Types.genericArrayType(Integer.class)));
+	}
+	
+	@Test
+	public void isPrimitiveWithParameterizedType()
+	{
+		assertFalse(TypeUtils.isPrimitive(Types.parameterizedType(List.class, Integer.class)));
+	}
+	
+	@Test
+	public void isPrimitiveWithWildcardType()
+	{
+		assertFalse(TypeUtils.isPrimitive(Types.unboundedWildcardType()));
+	}
+	
 	// getComponentType tests -------------------------------------------------
 	
 	@Test
@@ -1034,6 +1199,50 @@ public class TypeUtilsTest
 			TypeUtils.getArrayType(Types.unboundedWildcardType()));
 	}
 	
+	// isParameterizedType tests ----------------------------------------------
+	
+	@Test(expected = NullPointerException.class)
+	public void isParameterizedTypeWithNullType()
+	{
+		TypeUtils.isParameterizedType(null, List.class);
+	}
+	
+	@Test(expected = NullPointerException.class)
+	public void isParameterizedTypeWithNullRawType()
+	{
+		TypeUtils.isParameterizedType(List.class, null);
+	}
+	
+	@Test
+	public void isParameterizedTypeWithSameRawType()
+	{
+		assertTrue(TypeUtils.isParameterizedType(valueOf("List<Integer>"), List.class));
+	}
+	
+	@Test
+	public void isParameterizedTypeWithSuperRawType()
+	{
+		assertTrue(TypeUtils.isParameterizedType(valueOf("List<Integer>"), Collection.class));
+	}
+	
+	@Test
+	public void isParameterizedTypeWithUnassignableRawType()
+	{
+		assertFalse(TypeUtils.isParameterizedType(valueOf("List<Integer>"), Number.class));
+	}
+	
+	@Test
+	public void isParameterizedTypeWithClass()
+	{
+		assertFalse(TypeUtils.isParameterizedType(List.class, List.class));
+	}
+	
+	@Test
+	public void isParameterizedTypeWithParameterizedTypeArgument()
+	{
+		assertTrue(TypeUtils.isParameterizedType(valueOf("List<List<Integer>>"), List.class));
+	}
+	
 	// isSimpleParameterizedType tests ----------------------------------------
 	
 	@Test(expected = NullPointerException.class)
@@ -1092,6 +1301,48 @@ public class TypeUtilsTest
 		assertEquals(Integer.class, TypeUtils.getActualTypeArgument(valueOf("List<Integer>")));
 	}
 	
+	@Test(expected = IllegalArgumentException.class)
+	public void getActualTypeArgumentWithUnparameterizedType()
+	{
+		TypeUtils.getActualTypeArgument(Integer.class);
+	}
+	
+	@Test(expected = NullPointerException.class)
+	public void getActualTypeArgumentIndexWithNull()
+	{
+		TypeUtils.getActualTypeArgument(null, 0);
+	}
+	
+	@Test(expected = IllegalArgumentException.class)
+	public void getActualTypeArgumentIndexWithUnparameterizedType()
+	{
+		TypeUtils.getActualTypeArgument(Integer.class, 0);
+	}
+	
+	@Test
+	public void getActualTypeArgumentIndexWithFirstIndex()
+	{
+		assertEquals(Integer.class, TypeUtils.getActualTypeArgument(valueOf("List<Integer>"), 0));
+	}
+	
+	@Test
+	public void getActualTypeArgumentIndexWithSecondIndex()
+	{
+		assertEquals(Number.class, TypeUtils.getActualTypeArgument(valueOf("Map<Integer, Number>"), 1));
+	}
+	
+	@Test(expected = IllegalArgumentException.class)
+	public void getActualTypeArgumentIndexWithNegativeIndex()
+	{
+		TypeUtils.getActualTypeArgument(valueOf("List<Integer>"), -1);
+	}
+	
+	@Test(expected = IllegalArgumentException.class)
+	public void getActualTypeArgumentIndexWithOutOfBoundsIndex()
+	{
+		TypeUtils.getActualTypeArgument(valueOf("List<Integer>"), 1);
+	}
+	
 	// getResolvedSuperclass tests ---------------------------------------------
 	
 	@Test
@@ -1250,6 +1501,12 @@ public class TypeUtilsTest
 		assertEquals("java.lang.Integer[]", TypeUtils.toString(Integer[].class));
 	}
 	
+	@Test
+	public void toStringWithMultidimensionalArrayClass()
+	{
+		assertEquals("java.lang.Integer[][]", TypeUtils.toString(Integer[][].class));
+	}
+	
 	@Test
 	public void toStringWithUnboundedTypeVariable()
 	{
